#include <sys/ioctl.h>
#undef   ctrl()
#include <sys/time.h>

#include "defs.h"

/*
 *  Handle keyboard input.
 *
 *  Need to handle edit character mapping --- see if char is one
 *  of his edit characters, swap to standard character.
 */

static struct key {
	int		his;		/* his key */
	int		std;		/* what we'll map it to */
} keys [] = {
	{ ctrl(M),	ctrl(J) },		/* <cr> -> newline */
	{ 0, 		ctrl(H) },		/* backspace */
	{ 0, 		ctrl(W) },		/* word erase */
	{ 0, 		ctrl(X) },		/* line kill */
	{ 0, 		0       }
};


keyboard ()
{
	static   char   c, c2;              /* character read in    */
	register int 	k;				/* for doing key maps   */
	static   int    command = 0;	/* on bottom line?      */
	static   int    cpos = 0;		/* position in buffer   */
	static   char   cbuf[80];		/* and command buffer   */
	extern   int	childpid, tochild;
	extern Win      *outwin ;

	if (read (0, &c, 1) != 1)
		return;
	
	/*
	 *  We remap the basic editing characters here to make life easier later.
	 */

	for (k = 0; k < 4; k++) {
		if (keys[k].his == c) {
			c = keys[k].std;
			break;
		}
	}

	if (c == ctrl(L) || c == ctrl(R)) {
		w_redraw_screen() ;
		return;
	}

       c2 = 0 ;

	if (c == ESC) {				/* toggle command mode */
              int     mask ;
	      struct timeval  timeout ;
	      mask = (1 << 0) ;
	      timeout.tv_sec = 1 ;
	      timeout.tv_usec = 0 ;
	      if (!(select(32, &mask, 0, 0, &timeout) > 0 &&
		    read(0, &c2, 1) == 1 &&
		    (c2 == '$' || c2 == '(')))
              {
		      w_clear (msgwin);
		      command = !command; /* toggle mode */
		      if (command) /* entering command mode  */
			      w_putstr (msgwin, PROMPT, 0);
                       cpos = 0; /* and reset buffer index */
		       w_update (msgwin) ;
		       w_setpos (outwin, outwin->w_pos.y, outwin->w_pos.x);
		       /* restore cursor position */
		       w_update (outwin);
		       if (c2 == 0)
			       return ;
                      c = c2 ;
		      c2 = 0 ;
               }
        }

redo :
	if (command == 0) {						/* not in command mode */
              extern Win      *outwin ;

		if ((c == ctrl(D)) && childpid) 	/* send eof to child */
			close (tochild);
		if (childpid == 0 || Echo) {		/* send char to the socket */
#ifdef LOCAL_ECHO
			int old = selwin (0);
			showch (c);
			w_update (outwin) ;
			selwin (old);
                        w_update (outwin);
#endif LOCAL_ECHO
			(void) write (stream, &c, 1);
		}
		if (childpid && !Hold)				/* and to child process */
			(void) write (tochild, &c, 1);
	} else {

		/* 
		 *    The rest is command-mode processing where 
		 *    we fiddle with the command buffer. This 
		 *    should move to a separate routine somewhere.
		 */

		if (c == ctrl(J)) {			/* execute the buffer */
			cbuf[cpos] = '\0';		/* null-terminate it  */
			w_clear (msgwin);                    /* clear line to 
			end  */
			w_redraw (msgwin);                             /* upda
			te the screen  */
			if (cpos)
				execute (cbuf);		/* and execute it !!! */
			command = cpos = 0;		/* turn off command mode */
			w_setpos (outwin, outwin->w_pos.y, outwin->w_pos.x);
					      /* restore cursor position */
                         w_update (outwin);
			return;					/* and we're done */
		}

		/*
		 *  The message() routine modified the bottom line,
		 *  so reprint whatever should be down there.
		 */

		if (touched25) {
			w_setpos (msgwin, 0, 0);                 /* go to begi
			nning of line */
			w_putstr (msgwin, PROMPT, 0);                /* print 
			prompt */
			w_vputstr (msgwin, cbuf, cpos, 0);                  /*
			 and print buffer */
                         w_clrtoeol (msgwin);
			touched25 = 0; 
		}

                 w_setpos (msgwin, 0, strlen (PROMPT) + cpos);

		/*
		 *  If we're here, then we want to add the character to
		 *  the pending command buffer.
		 */

		switch (c) {
			case ctrl(H):	/* backspace */
							if (cpos) {
								cpos--;
							      w_deletech(msgwin);
							}
							break;

			case ctrl(W):	/* word erase */
							while (--cpos >= 0) {
								if (cbuf[cpos] != ' ') {
									cpos++;
									break;
								}
							w_deletech(msgwin) ;
							}
							while (--cpos >= 0) {
								if (cbuf[cpos] == ' ') {
									cpos++;
									break;
								}
							w_deletech(msgwin) ;
							}
							break;

			case ctrl(U):	/* line kill */
                        case ctrl(X):   w_setpos (msgwin, 0, strlen (PROMPT));
							w_clrtoeol (msgwin);
							cpos = 0;
							break;

			case ctrl(I):	/* tab */
							c = ' ';     /* fall through ... */
			default:		cbuf[cpos++] = c;
							w_putstr (msgwin, &c,0);
							break;
		}
		w_update (msgwin) ;
		cbuf[cpos] = '\0';
		w_setpos (outwin, outwin->w_pos.y, outwin->w_pos.x) ;
				      /* restore cursor position */
                w_update (outwin);
	}

       if ( c2 != 0 ) {
	       c = c2 ;
	       c2 = 0 ;
	       goto  redo ;
	}
}


/*
 *   Get our editing characters. Called by setup().
 */

geteditchars ()
{
	struct sgttyb  chars;		/* standard char- and line-erase */
#ifdef TIOCGLTC
	struct ltchars lchars;		/* local - word-erase */
#endif

	if (ioctl (0, TIOCGETP, &chars) < 0) {
		error (0, "ioctl (TIOCGETP) failed");
		return;
	}
	keys[1].his = chars.sg_erase;
	keys[3].his = chars.sg_kill;

#ifdef TIOCGLTC
	if (ioctl (0, TIOCGLTC, &lchars) < 0) {
		error (0, "ioctl (TIOCGLTC) failed");
		return;
	}
	keys[2].his = lchars.t_werasc;
#endif
}
