/*
 *   Routines to manage the windows.  SOme of this is pretty ugly
 *   stuff, especially as most of the time is spent here.
 *
 *   We manage to resize windows by buffering all parts of a conversation,
 *   so we just (re)print the buffered part.  The buffering also accounts
 *   for most of the lack of speed.  A future enhancement will probably
 *   be to replace the constant buffering with a smarter routine that
 *   will just grab each window from the curses screen buffer when we call
 *   stretch() ...
 */


#include <sys/ioctl.h>
#include <signal.h>
#include "defs.h"

#ifndef WINDOWS
#define WINDOWS 32                /* we can only fit 23 on a sun !!! */
#endif

struct window {
	Win   *win ;          /* main window		    */
	Win   *twin ;         /* title window		    */
        int   inuse;          /* 1 if in use, 0 if not      */
        char  *login;         /* login name of this person  */
        char  *host;          /* their host name            */
        char  *tty;           /* tty                        */
        char  *realname;      /* finger name, if any        */
        int   old;            /* index of oldest ch in buf  */
        int   new;            /* where next char goes       */
        char  *buf;           /* text buffer                */
	char	*linebuf ;	/* line buffer for kana conversion */
	char	*wordbeg ;	/* begining of current word */
	char	*wordend ;	/* end of current word */
	char	inquote ;	/* no conversion take place */
	char	inkanji ;	/* input kanji characters */
	char	escape ;	/* status reading escape sequence */	
	char	kanji1 ;	/* first byte of kanji */
};

#define reset_line(w)	((w)->wordbeg = (w)->wordend = (w)->linebuf)

static  struct window windows[WINDOWS];
static	struct window *currwin;    /* pointer to current window    */
Win		*outwin;	/* = currwin->win ; this variable is exported */
int             current = 0;                    /* currently active window      */
int             users;                                  /* number of users/windows      */
int             lower_y;
int             stretching = 0;
int             did_screen = 0;                 /* have set up curses          */
static  int y , x;                              /* current screen coordinates  */
extern  int sigstop();                  /* ^Z handling routine         */

static int ourwin;                              /* the window # that is _ours_ */

/*
 *  Mark all of the windows as not being in use.
 */

setup ()
{
        int  i;

	reset_attr_after_redraw = 1 ;
        w_init() ;

        for (i = 0; i < WINDOWS; i++) {
                windows[i].inuse = 0;
                windows[i].old   = 0;
                windows[i].new   = 0;
        }
        
        did_screen = 1;

        signal (SIGTSTP, sigstop);
        geteditchars ();                                /* get and save his editing characters */

        maxx = _screen->w_cols - 1;                       /* save max screen coordinates */
        maxy = _screen->w_lines - 1;

        lower_y = maxy - 2;
        y = 0;                                           /* initial window position */
        x = 0;
	windows[0].win = w_create(maxy - 1, 0, 1, 0) ;
	windows[0].twin = w_create(1, 0, 0, 0) ;
        currwin = &windows[0];
	outwin = currwin->win;
        w_setpos (windows[0].win, 0, 0);
	msgwin = w_create(1, 0, maxy, 0) ;
        users = 0;                                                                      /* not talking to anyone yet */
}


/*
 *  Restore everything to how it was before we came...
 */

cleanup ()
{
        w_end_screen() ;
        did_screen = 0;
}


/*
 *  Select the new current window to use, set the window bounds,
 *  and set the new cursor position ...  Returns the old window number.
 */

selwin (win)
char    win;
{
        int  old = current;

        if (win == current && !stretching)      /* no change */
                return (win);
        if (win == ourwin)                              /* map our window to window 0 */
                win = 0;

        if (windows[win].inuse == 0)
                return (old);

        if (win >= 0 && win < WINDOWS) {
#ifdef notdef
                currwin->y = y;                         /* save last coordinates */
                currwin->x = x;
                currwin = &windows[win];        /* set pointer to new window */
                y = currwin->y;                         /* and retrieve from new win */
                x = currwin->x;
#endif
                currwin = &windows[win];        /* set pointer to new window */
		outwin = currwin->win ;
                lower_y = outwin->w_lines ;
		y = outwin->w_pos.y ;
		x = outwin->w_pos.x ;
                w_setpos (outwin, y, x);          /* restore cursor pos */
                current = win;
        }
        return (old);
}



/*
 *  Add a new window to the screen.
 *  Buf looks like "user:host:tty:realname"
 */

adduser (slot, buf)
int     slot;
char    *buf;
{
        char    *i;

        /*
         *  We rely on the fact that the first window given to us is
         *  our own, so we remember which one that is and remap it 
         *  to the top window. This makes people happier.
         */

        if (users == 0) {
                ourwin = slot;
                slot = 0;
        }

        if (windows[slot].inuse)                                /* that window is already in use */
                return;

        windows[slot].inuse = 1;                                /* mark as being in use */
        windows[slot].buf = malloc (BUFFER);    /* and grab a buffer for it */
        windows[slot].old = 0;                                  /* nothing in the buffer yet */
        windows[slot].new = 0;
	windows[slot].linebuf = malloc (BUFFER);
	windows[slot].wordbeg = windows[slot].wordend = windows[slot].linebuf ;

        /*  get the info out of the buffer line  */
        if (i = index (buf, ':'))    /* name */
                *i++ = '\0';
        else
                return;
        windows[slot].login = strsave (buf);
        buf = i;
        i = index (buf, ':');    /* host */
        *i++ = '\0';
        windows[slot].host = strsave (buf);
        buf = i;
        i = index (buf, ':');    /* tty */
                *i++ = '\0';
        windows[slot].tty = strsave (buf);
        buf = i;
        windows[slot].realname = strsave (buf);  /* real name */

        if (users++ == 1)                                       /* add another user to the count */
                (void) write (1, "\07\07", 2);  /* initial connection - beep the user */
        stretch (0);                                            /* and recompute all the windows */
}



/*
 *  Delete a user and his window.
 */

deluser (win)
char     win;
{
        if (win < 0 || win >= WINDOWS)
                return;

        if (windows[win].inuse == 0)
                return;

        windows[win].inuse = 0;                 /* this window no longer in use */
        free (windows[win].login);              /* so free up any space allocated for it */
        free (windows[win].host);
        free (windows[win].tty);
        free (windows[win].realname);
        free (windows[win].buf);
	w_destroy(windows[win].win) ;
	w_destroy(windows[win].twin) ;

        if (--users > 1) {                              /* decrement the count */
                stretch (0);                            /* and redraw */
/*                w_redraw (scr);*/
        } else                                                  /* if only one user left .. */
                quit ();                                        /* then we're done */
}



/*
 *  Show the given character, performing word and line erase.
 */

showch (ch)
char    ch;
{
	register struct window	*w ;

        if (users == 0)
                return;

	w = currwin ;

        if (!stretching) {
                /* save char into buffer */
                if (w->new >= BUFFER)
                        error (1, "New >= BUFFER in showch()");
                w->buf[w->new++] = ch;
                if (w->new >= BUFFER)                     /* wrap around */
                        w->new = 0;
                if (w->new == w->old)       /* tail caught up */
                        w->old++;
                if (w->old >= BUFFER)                     /* wrap old around */
                        w->old = 0;
        }

	switch ( w->escape ) {
	case '\033' :
		if ( ch == '$' || ch == '(' )
			w->escape = ch ;
		else
			w->escape = 0 ;
		return ;

	case '$' :
		if ( ch == '@' || ch == 'B' )
			w->inkanji = 1 ;
		w->escape = 0 ;
		return ;

	case '(' :
		if ( ch == 'B' || ch == 'J' || ch == 'H' )
			w->inkanji = 0 ;
		w->escape = 0 ;
		return ;
	}

        switch (ch) {
	case '\033':
		if ( kanji_code >= 0 ) {
			w->escape = ch ;
			return ;
		}
		break ;

	case '\0177':		/* delete */
	case ctrl(H):		/* backspace */
		if (x > 0) {
			if ( w->wordend > w->wordbeg ) {
				/* in a word */
				w->wordend-- ;
				show_word(w, 0) ;
			} else {
				w_deletech(w->win) ;
				--x ;
				if ( *--w->wordend == '"' )
					w->inquote = !w->inquote ;
				w->wordbeg-- ;
				/* if back to just after a word, reconvert it */
				if ( !w->inquote ) {
					while ( w->wordbeg > w->linebuf ) {
						ch = *--w->wordbeg ;
						if ( !iswordch(ch) ) {
							++w->wordbeg ;
							break ;
						}
					}
					if ( w->wordend > w->wordbeg )
						show_word(w, 0) ;
				}
			}
		}
		break;

	case ctrl(I):		/* tab */
		add_ch(w, ' ') ;
		while (x & 7)
			add_ch(w, ' ') ;
		break;

	case ctrl(J):		/* newline */
		if ( w->wordend > w->wordbeg )
			show_word(w, 1) ;
		new_line(w) ;
		break;

	case ctrl(L):		/* clear to end of current line */
		w_clrtoeol (w->win);
		break;

	case ctrl(M):		/* carriage return */
		w_setpos (w->win, y, x = 0);
		reset_line(w) ;
		break;

	case ctrl(N):		/* clear window */
		w_clear(w->win) ;
		w_setpos (w->win, y = 0, x = 0);
		reset_line(w) ;
		break;

	case ctrl(U):		/* line kill */
	case ctrl(X):
		w_setpos (w->win, y, x = 0);
		w_clrtoeol (w->win);
		while ( w->wordend > w->linebuf ) {
			if ( *--w->wordend == '"' )
				w->inquote = !w->inquote ;
		}
		reset_line(w) ;
		break;

	case ctrl(W):		/* word erase */
		while (x > 0) { /* nuke the spaces */
			if (*--w->wordend != ' ') {
				++w->wordend ;
				break;
			}
			--x ;
		}
		while ( w->wordend > w->linebuf ) {	 /* and the non-spaces */
			if (*--w->wordend == '"')
				w->inquote = !w->inquote ;
			if (*w->wordend == ' ') {
				++w->wordend ;
				break ;
			}
		}
		w->wordbeg = w->wordend ;
		w_setpos(w->win, y, x) ;
		while ( x > 0 ) {
			if ( wc_char(w_getch(w->win, y, --x)) == ' ' ) {
				++x ;
				break ;
			}
			w_deletech(w->win) ;
		}
		break;

	case ctrl(^):		/* home cursor */
		w_setpos (w->win, y = 0, x = 0);
		reset_line(w) ;
		break;

	case ctrl(G):		/* bell - switchable */
		if (Bells) {
			w_bell() ;
			break;
		}
		/* fall through ... */

	default:
		if ( w->inkanji ) {
			if ( w->kanji1 ) {
				*w->wordend++ = w->kanji1 | 0x80 ;
				*w->wordend++ = ch | 0x80 ;
				show_word(w, 0) ;
				w->kanji1 = 0 ;
			} else
				w->kanji1 = ch ;
		} else if ( w->inquote || !iswordch(ch) ) {
			if (ch < ' ') {
				add_ch(w, '^') ;
				ch += '@';
			}
			add_ch(w, ch) ;
			if (ch == '"')
				w->inquote = !w->inquote ;
		} else {
			*w->wordend++ = ch ;
			show_word(w, 0) ;
		}
		break;
        }
}

static
new_line(w)
	struct window	*w ;
{
	if (++y >= lower_y)
		y = 0;
	w_setpos(w->win, y, x = 0);
	w_clrtoeol (w->win);
	reset_line(w) ;
}

static
add_ch(w, ch)
	register struct window	*w ;
	char	ch ;
{
	if ( w->wordend > w->wordbeg ) {
		show_word(w, 1) ;	/* fix the word conversion */
		w->wordbeg = w->wordend ;
	}
	*w->wordbeg++ = ch ;
	w->wordend = w->wordbeg ;
	w_vputstr(w->win, &ch, 1, 0) ;
	if ( ++x >= maxx ) {
		new_line(w) ;
	}
}

static
show_word(w, complete)
	struct window	*w ;
	int	complete ;
{
	register int	wx ;
	int	l ;
	char	*s ;
	char	ch ;
	char	coutbuf[BUFFER] ;

	coutbuf[0] = 0 ;
	if ( kanji_code < 0 ||
	     convert_word(w->wordbeg, w->wordend, coutbuf, complete) )
	{
		l = w->wordend - w->wordbeg ;
		s = w->wordbeg ;
	} else {
		l = strlen(coutbuf) ;
		s = coutbuf ;
	}
	for ( wx = x - 1 ; wx >= 0 ; wx-- ) {
		ch = wc_char(w_getch(w->win, y, wx)) ;
		if ( !iswordch(ch) )
			break ;
		w_deletech(w->win) ;
	}
	++wx ;
	if ( (x = wx + l) > maxx ) {
		if ( wx == 0 ) {
			/* this might not be correct */
			reset_line(w) ;
			w_vputstr(w->win, s, l, 0) ;
			x = 0 ;
			if ( ++y >= lower_y )
				y = 0 ;
			return ;
		} else {
			char	*bp = w->wordbeg ;
			char	*ep = w->wordend ;
			new_line(w) ;
			while ( bp < ep ) {
				*w->wordend++ = *bp++ ;
			}
			x = l ;
		}
	}
	w_vputstr(w->win, s, l, 0) ;
}

/*
 *  Print a message to the bottom line.
 */

putmessage (str)
char *str;
{
        extern int Inverse;
        extern int did_screen;
	Wchar	attr ;

        if (!did_screen) {
                printf ("%s\r\n", str);
                fflush (stdout);
                return;
        }
        touched25 = 1;                          /* has been changed  */
        w_clear (msgwin);                            /* clear it          */
        w_redraw (msgwin);                                     /* they see it clear */
	attr = Inverse? A_REVERSE : 0 ;

        while (*str) {                          /* print out the message, removing ctrl chars */
		static char	ctrlstr[2] = "^" ;
                if (*str < ' ') {
			ctrlstr[1] = *str + '@' ;
                        w_vputstr (msgwin, ctrlstr, 2, attr) ;
                } else
                        w_vputstr (msgwin, str, 1, attr);
                str++;
        }
	w_update (msgwin) ;

        w_setpos (outwin, y, x);			/* restore old pos   */
        w_update (outwin);                            /* etc ...           */
}


/*
 *  Print out a header for the given slot.
 */

header (num)
int num;
{
        register int i;
        extern   int Inverse;
	char	*bp ;

        if (windows[num].inuse == 0) {
                error (0, "header(): slot not in use");
                return;
        }

        sprintf (buf, "---- %s@%s on %s (", windows[num].login, 
                        windows[num].host, windows[num].tty) ;
	bp = buf + strlen(buf) ;
	if ( kanji_code < 0 )
		strcpy(bp, windows[num].realname) ;
	else
		convert_string(windows[num].realname, bp) ;
	bp += strlen(bp) ;
	*bp++ = ')' ;
	*bp++ = ' ' ;
        
        for (i = bp - buf; i < maxx; i++)
                *bp++ = '-';
        buf[maxx-2] = '\0';             /* two (four?) for magic-cookie terminals */

        w_setpos (windows[num].twin, 0, 0);
        w_putstr (windows[num].twin, buf, Inverse? A_REVERSE: 0);
	w_update (windows[num].twin) ;
}



/*
 *  Take the current screen and contort it to handle the new windows.
 *
 *  Handles both the screen changing size and just the 
 *  number of windows changing.
 *
 *  We set caught_sig to be non-zero if coming here on SIGWINCH, 0 otherwise.
 *  We also only refresh the screen if we caught the signal....
 */

stretch (caught_sig)
int caught_sig;
{
        register int   i, w;
        register int   extra;
        register int   lines;
        int      oldy, oldx;
        int      newy, newx;
        int      lines_win;
        int      oldcurr = current;

        if (users == 0)         /* nothing to recompute */
                return;

        stretching = 1;

        oldy = maxy + 1;
        oldx = maxx + 1;

        if (caught_sig) {                       /* got SIGWINCH, have to redo screen */
#if defined(TIOCGWINSZ) || defined(TIOCGSIZE)
#ifdef TIOCGWINSZ
                struct winsize winsize;
#else  TIOCGSIZE        /* (older?) suns */
                struct ttysize ttysize;
#endif TIOCGSIZE

                /* get new screen size */
#ifdef TIOCGWINSZ
                if (ioctl (0, TIOCGWINSZ, &winsize)) {
                        error (0, "ioctl: TIOCGWINSZ");
                        return;
                }
                newy = winsize.ws_row;
                newx = winsize.ws_col;
#else TIOCGSIZE
                if (ioctl (0, TIOCGSIZE, &ttysize)) {
                        error (0, "ioctl: TIOCGSIZE");
                        return;
                }
                newy = ttysize.ts_lines;
                newx = ttysize.ts_cols;
#endif TIOCGSIZE 
                
#else /* don't have window size ioctls */
                /* 
                 *  Can't get new window size, so presumably sigwinch 
                 *  doesn't exist and this code will never happen !!!
                 */
                newy = oldy;
                newx = oldx;
#endif TIOCGWINSZ
                maxy = newy - 1;        /* max screen coordinate - zero-based */
                maxx = newx - 1;
        } else {                                /* screen size hasn't changed */
                newy = oldy;
                newx = oldx;
        }

        if (newx != oldx || newy != oldy) {                     /* have to shut down curses */
		w_end() ;
		w_destroy(msgwin) ;
                did_screen = 0;

                w_init ();                     /* and restart it */
		msgwin = w_create(1, 0, maxy, 0) ;
                did_screen = 1;
                signal (SIGTSTP, sigstop);
        } else {                                /* just have to clear all the lines */
/*
		for (i = 0; i < newy; i++) {
                        w_setpos (scr, i, 0);
                        w_clrtoeol (scr);
                }
*/
      }

        lines_win = (newy - 1) / users;         /* recompute window boundaries */
        extra = (newy - 1) - (users * lines_win);       /* total extra lines */
        lines = 0;

        /* 
         *   Now calculate the new boundaries and
         *   copy the buffer back to the new window.
         */
        
/*        x = 0; y = 0;           /* go back to the beginning of the window */
        for (w = 0; w < WINDOWS; w++) {         /* now go copy boundaries */
                if (windows[w].inuse == 0)
                        continue;
		if ( windows[w].win ) {
			w_destroy(windows[w].win) ;
			w_destroy(windows[w].twin) ;
		}
		windows[w].twin = w_create(1, 0, lines, 0) ;
                if (extra-- > 0) {                       /* extra lines for this window */
			windows[w].win = w_create(lines_win, 0, lines + 1, 0) ;
                        lines++;
		} else
			windows[w].win = w_create(lines_win - 1, 0, lines + 1, 0) ;
                lines += lines_win;                     /* add on lines */
                header (w);                                     /* redraw the window header */
                selwin (w);                                     /* switch to this window */
                /* and dump the buffer back into the window */
                for (i = windows[w].old; i != windows[w].new; i = (i+1) % BUFFER)
                        showch (windows[w].buf[i]);
		w_update(windows[w].win) ;
        }

        selwin (oldcurr);                               /* reselect the right window */
	w_update (outwin) ;
/*        if (caught_sig)
                w_redraw (scr);                                     /* and make it all visible */
        stretching = 0;
}
